"""Studio Web File Handler for managing file operations in UiPath projects."""

import json
import logging
import os
from datetime import datetime, timezone
from typing import Any, AsyncIterator, Dict, Optional, Set

from ...models.exceptions import EnrichedException
from .._utils._constants import (
    AGENT_INITIAL_CODE_VERSION,
    AGENT_STORAGE_VERSION,
    AGENT_TARGET_RUNTIME,
    AGENT_VERSION,
)
from .._utils._project_files import (  # type: ignore
    FileInfo,
    FileOperationUpdate,
    files_to_include,
    read_toml_project,
)
from .._utils._studio_project import (
    AddedResource,
    ModifiedResource,
    ProjectFile,
    ProjectFolder,
    ProjectStructure,
    StructuralMigration,
    StudioClient,
)

logger = logging.getLogger(__name__)


class SwFileHandler:
    """Handler for Studio Web file operations.

    This class encapsulates all file operations for UiPath Studio Web projects,
    including uploading, updating, deleting, and managing project structure.

    Attributes:
        directory: Local project directory
        include_uv_lock: Whether to include uv.lock file
    """

    def __init__(
        self,
        project_id: str,
        directory: str,
        include_uv_lock: bool = True,
    ) -> None:
        """Initialize the SwFileHandler.

        Args:
            project_id: The ID of the UiPath project
            directory: Local project directory
            include_uv_lock: Whether to include uv.lock file
        """
        self.directory = directory
        self.include_uv_lock = include_uv_lock
        self._studio_client = StudioClient(project_id)
        self._project_structure: Optional[ProjectStructure] = None

    def _get_folder_by_name(
        self, structure: ProjectStructure, folder_name: str
    ) -> Optional[ProjectFolder]:
        """Get a folder from the project structure by name.

        Args:
            structure: The project structure
            folder_name: Name of the folder to find

        Returns:
            Optional[ProjectFolder]: The found folder or None
        """
        for folder in structure.folders:
            if folder.name == folder_name:
                return folder
        return None

    def collect_all_files(
        self,
        folder: ProjectFolder,
        files_dict: Dict[str, ProjectFile],
        current_path: str = "",
    ) -> None:
        """Recursively collect all files from a folder with computed paths.

        Args:
            folder: The folder to traverse
            files_dict: Dictionary to store files (indexed by name)
            current_path: The current path prefix for files in this folder
        """
        # Add files from current folder
        for file in folder.files:
            file_path = f"{current_path}/{file.name}" if current_path else file.name
            files_dict[file_path] = file

        # Recursively process subfolders
        for subfolder in folder.folders:
            subfolder_path = (
                f"{current_path}/{subfolder.name}" if current_path else subfolder.name
            )
            self.collect_all_files(subfolder, files_dict, subfolder_path)

    def _get_remote_files(
        self,
        structure: ProjectStructure,
        source_code_folder: Optional[ProjectFolder] = None,
    ) -> tuple[Dict[str, ProjectFile], Dict[str, ProjectFile]]:
        """Get all files from the project structure indexed by name.

        Args:
            structure: The project structure
            source_code_folder: Optional source_code folder to collect files from

        Returns:
            Tuple of (root_files, source_code_files) dictionaries with file paths as keys
        """
        root_files: Dict[str, ProjectFile] = {}
        source_code_files: Dict[str, ProjectFile] = {}

        # Add files from root level
        for file in structure.files:
            root_files[file.name] = file

        # Add files from source_code folder if it exists
        if source_code_folder:
            self.collect_all_files(source_code_folder, source_code_files)

        return root_files, source_code_files

    async def _process_file_uploads(
        self,
        local_files: list[FileInfo],
        source_code_files: Dict[str, ProjectFile],
        root_files: Dict[str, ProjectFile],
    ) -> list[FileOperationUpdate]:
        """Process all file uploads to the source_code folder.

        This method:
        1. Compares local files with remote files
        2. Builds a structural migration with added/modified/deleted resources
        3. Prepares agent.json and entry-points.json
        4. Performs the structural migration
        5. Cleans up empty folders

        Args:
            local_files: List of files to upload
            source_code_files: Dictionary of existing remote files
            root_files: Dictionary of existing root-level files

        Returns:
            List of FileOperationUpdate objects describing all file operations

        Raises:
            Exception: If any file upload fails
        """
        structural_migration = StructuralMigration(
            deleted_resources=[], added_resources=[], modified_resources=[]
        )
        processed_source_files: Set[str] = set()
        updates: list[FileOperationUpdate] = []

        # Process each local file and build structural migration
        for local_file in local_files:
            if not os.path.exists(local_file.file_path):
                logger.info(f"File not found: '{local_file.file_path}'")
                continue

            # Skip agent.json as it's handled separately
            if local_file.file_name == "agent.json":
                continue

            remote_file = source_code_files.get(
                local_file.relative_path.replace("\\", "/"), None
            )
            if remote_file:
                # File exists remotely - mark for update
                processed_source_files.add(remote_file.id)
                structural_migration.modified_resources.append(
                    ModifiedResource(
                        id=remote_file.id, content_file_path=local_file.file_path
                    )
                )
                updates.append(
                    FileOperationUpdate(
                        file_path=local_file.file_name,
                        status="updating",
                        message=f"Updating '{local_file.file_name}'",
                    )
                )
            else:
                # File doesn't exist remotely - mark for upload
                parent_path = os.path.dirname(local_file.relative_path)
                structural_migration.added_resources.append(
                    AddedResource(
                        content_file_path=local_file.file_path,
                        parent_path=f"source_code/{parent_path}"
                        if parent_path != ""
                        else "source_code",
                    )
                )
                updates.append(
                    FileOperationUpdate(
                        file_path=local_file.relative_path,
                        status="uploading",
                        message=f"Uploading '{local_file.relative_path}'",
                    )
                )

        # Identify and add deleted files (files that exist remotely but not locally)
        deleted_files = self._collect_deleted_files(
            source_code_files, processed_source_files
        )
        structural_migration.deleted_resources.extend(deleted_files)

        # Add delete updates
        for file_id in deleted_files:
            file_name = next(
                (name for name, f in source_code_files.items() if f.id == file_id),
                file_id,
            )
            updates.append(
                FileOperationUpdate(
                    file_path=file_name,
                    status="deleting",
                    message=f"Deleting '{file_name}'",
                )
            )

        # Load uipath.json configuration
        with open(os.path.join(self.directory, "uipath.json"), "r") as f:
            uipath_config = json.load(f)

        # Prepare agent.json migration (may download existing file to increment version)
        agent_update = await self._prepare_agent_json_migration(
            structural_migration, root_files, uipath_config
        )
        if agent_update:
            updates.append(agent_update)

        # Prepare entry-points.json migration (may download existing file to merge)
        entry_points_update = await self._prepare_entrypoints_json_migration(
            structural_migration, root_files, uipath_config
        )
        if entry_points_update:
            updates.append(entry_points_update)

        # Perform the structural migration (uploads/updates/deletes all files)
        await self._studio_client.perform_structural_migration_async(
            structural_migration
        )

        # Clean up empty folders after migration
        await self._cleanup_empty_folders()

        return updates

    def _collect_deleted_files(
        self,
        source_code_files: Dict[str, ProjectFile],
        processed_source_file_ids: Set[str],
    ) -> set[str]:
        """Identify remote files that no longer exist locally.

        Args:
            source_code_files: Dictionary of existing remote files
            processed_source_file_ids: Set of file IDs that were processed (exist locally)

        Returns:
            Set of file IDs to delete
        """
        deleted_file_ids: Set[str] = set()

        for _, remote_file in source_code_files.items():
            if remote_file.id not in processed_source_file_ids:
                deleted_file_ids.add(remote_file.id)

        return deleted_file_ids

    async def _cleanup_empty_folders(self) -> None:
        """Delete empty folders from the project structure.

        This method:
        1. Gets the current project structure
        2. Recursively finds all empty folders
        3. Deletes each empty folder
        """
        structure = await self._studio_client.get_project_structure_async()
        source_code_folder = self._get_folder_by_name(structure, "source_code")

        if not source_code_folder:
            return

        empty_folders = self._find_empty_folders(source_code_folder)

        if empty_folders:
            for folder in empty_folders:
                await self._studio_client.delete_item_async(folder["id"])
                logger.info(f"Deleted empty folder: '{folder['name']}'")

    def _find_empty_folders(self, folder: ProjectFolder) -> list[dict[str, str]]:
        """Recursively find all empty folders.

        Args:
            folder: The folder to check

        Returns:
            List of empty folder info dictionaries with 'id' and 'name' keys
        """
        empty_folders: list[dict[str, str]] = []

        for subfolder in folder.folders:
            # Recursively check subfolders first
            empty_folders.extend(self._find_empty_folders(subfolder))

            # Check if current subfolder is empty after processing its children
            if self._is_folder_empty(subfolder):
                if subfolder.id is not None:
                    empty_folders.append({"id": subfolder.id, "name": subfolder.name})

        return empty_folders

    def _is_folder_empty(self, folder: ProjectFolder) -> bool:
        """Check if a folder is empty (no files and no non-empty subfolders).

        Args:
            folder: The folder to check

        Returns:
            True if the folder is empty, False otherwise
        """
        if folder.files:
            return False

        if not folder.folders:
            return True

        # If folder has subfolders, check if all subfolders are empty
        for subfolder in folder.folders:
            if not self._is_folder_empty(subfolder):
                return False

        return True

    async def _prepare_entrypoints_json_migration(
        self,
        structural_migration: StructuralMigration,
        root_files: Dict[str, ProjectFile],
        uipath_config: Dict[str, Any],
    ) -> Optional[FileOperationUpdate]:
        """Prepare entry-points.json to be included in the same structural migration.

        This method:
        1. Downloads existing entry-points.json if it exists
        2. Merges entryPoints from uipath.json config
        3. Adds to structural migration as modified or added resource

        Args:
            structural_migration: The structural migration to add resources to
            root_files: Dictionary of root-level files
            uipath_config: Configuration from uipath.json

        Returns:
            FileOperationUpdate describing the operation, or None if error occurred
        """
        existing = root_files.get("entry-points.json")

        if existing:
            # Entry-points.json exists - download and merge
            try:
                entry_points_json = (
                    await self._studio_client.download_project_file_async(existing)
                ).json()
                entry_points_json["entryPoints"] = uipath_config["entryPoints"]
            except Exception:
                logger.info(
                    "Could not parse existing 'entry-points.json' file, using default version"
                )
                # If parsing fails, create default structure
                entry_points_json = {
                    "$schema": "https://cloud.uipath.com/draft/2024-12/entry-point",
                    "$id": "entry-points.json",
                    "entryPoints": uipath_config["entryPoints"],
                }

            structural_migration.modified_resources.append(
                ModifiedResource(
                    id=existing.id,
                    content_string=json.dumps(entry_points_json),
                )
            )
            return FileOperationUpdate(
                file_path="entry-points.json",
                status="updating",
                message="Updating 'entry-points.json'",
            )
        else:
            # Entry-points.json doesn't exist - create new one
            logger.info(
                "'entry-points.json' file does not exist in Studio Web project, initializing using default version"
            )
            entry_points_json = {
                "$schema": "https://cloud.uipath.com/draft/2024-12/entry-point",
                "$id": "entry-points.json",
                "entryPoints": uipath_config["entryPoints"],
            }
            structural_migration.added_resources.append(
                AddedResource(
                    file_name="entry-points.json",
                    content_string=json.dumps(entry_points_json),
                )
            )
            return FileOperationUpdate(
                file_path="entry-points.json",
                status="uploading",
                message="Uploading 'entry-points.json'",
            )

    async def _prepare_agent_json_migration(
        self,
        structural_migration: StructuralMigration,
        root_files: Dict[str, ProjectFile],
        uipath_config: Dict[str, Any],
    ) -> Optional[FileOperationUpdate]:
        """Prepare agent.json to be included in the same structural migration.

        This method:
        1. Extracts author from JWT token or pyproject.toml
        2. Downloads existing agent.json if it exists to increment code version
        3. Builds complete agent.json structure
        4. Adds to structural migration as modified or added resource

        Args:
            structural_migration: The structural migration to add resources to
            root_files: Dictionary of root-level files
            uipath_config: Configuration from uipath.json

        Returns:
            FileOperationUpdate describing the operation, or None if error occurred
        """

        def get_author_from_token_or_toml() -> str:
            """Get author from JWT token or fall back to pyproject.toml."""
            import jwt

            token = os.getenv("UIPATH_ACCESS_TOKEN")
            if token:
                try:
                    decoded_token = jwt.decode(
                        token, options={"verify_signature": False}
                    )
                    preferred_username = decoded_token.get("preferred_username")
                    if preferred_username:
                        return preferred_username
                except Exception:
                    # If JWT decoding fails, fall back to toml
                    pass

            toml_data = read_toml_project(
                os.path.join(self.directory, "pyproject.toml")
            )
            return toml_data.get("authors", "").strip()

        # Extract input and output schemas from entrypoints
        try:
            input_schema = uipath_config["entryPoints"][0]["input"]
            output_schema = uipath_config["entryPoints"][0]["output"]
        except (FileNotFoundError, KeyError) as e:
            logger.error(
                f"Unable to extract entrypoints from configuration file. Please run 'uipath init' : {str(e)}",
            )
            return None

        author = get_author_from_token_or_toml()

        # Initialize agent.json structure with metadata
        agent_json = {
            "version": AGENT_VERSION,
            "metadata": {
                "storageVersion": AGENT_STORAGE_VERSION,
                "targetRuntime": AGENT_TARGET_RUNTIME,
                "isConversational": False,
                "codeVersion": AGENT_INITIAL_CODE_VERSION,
                "author": author,
                "pushDate": datetime.now(timezone.utc).isoformat(),
            },
            "inputSchema": input_schema,
            "outputSchema": output_schema,
            "bindings": uipath_config.get(
                "bindings", {"version": "2.0", "resources": []}
            ),
            "settings": {},
            # TODO: remove this after validation check gets removed on SW side
            "entryPoints": [{}],
        }

        existing = root_files.get("agent.json")
        if existing:
            # Agent.json exists - download and increment version
            try:
                existing_agent_json = (
                    await self._studio_client.download_project_file_async(existing)
                ).json()
                version_parts = existing_agent_json["metadata"]["codeVersion"].split(
                    "."
                )
                if len(version_parts) >= 3:
                    # Increment patch version (0.1.0 -> 0.1.1)
                    version_parts[-1] = str(int(version_parts[-1]) + 1)
                    agent_json["metadata"]["codeVersion"] = ".".join(version_parts)
                else:
                    # Invalid version format, use default with patch = 1
                    agent_json["metadata"]["codeVersion"] = (
                        AGENT_INITIAL_CODE_VERSION[:-1] + "1"
                    )
            except Exception:
                logger.info(
                    "Could not parse existing 'agent.json' file, using default version"
                )

            structural_migration.modified_resources.append(
                ModifiedResource(
                    id=existing.id,
                    content_string=json.dumps(agent_json),
                )
            )
            return FileOperationUpdate(
                file_path="agent.json",
                status="updating",
                message="Updating 'agent.json'",
            )
        else:
            # Agent.json doesn't exist - create new one
            logger.info(
                "'agent.json' file does not exist in Studio Web project, initializing using default version"
            )
            structural_migration.added_resources.append(
                AddedResource(
                    file_name="agent.json",
                    content_string=json.dumps(agent_json),
                )
            )
            return FileOperationUpdate(
                file_path="agent.json",
                status="uploading",
                message="Uploading 'agent.json'",
            )

    async def upload_source_files(
        self, settings: Optional[dict[str, Any]]
    ) -> AsyncIterator[FileOperationUpdate]:
        """Main method to upload source files to the UiPath project.

        This method:
        1. Gets project structure (or creates if doesn't exist)
        2. Creates source_code folder if needed
        3. Collects local files to upload
        4. Processes file uploads (yields progress updates)
        5. Performs structural migration
        6. Cleans up empty folders

        Args:
            settings: File handling settings (includes/excludes)

        Yields:
            FileOperationUpdate: Progress updates for each file operation

        Raises:
            Exception: If any step in the process fails
        """
        # Get or create project structure
        try:
            structure = await self._studio_client.get_project_structure_async()
        except EnrichedException as e:
            if e.status_code == 404:
                # Project structure doesn't exist - create empty structure and lock
                structure = ProjectStructure(name="", files=[], folders=[])
                await self._studio_client._put_lock()
            else:
                raise

        source_code_folder = self._get_folder_by_name(structure, "source_code")
        root_files, source_code_files = self._get_remote_files(
            structure, source_code_folder
        )

        # Create source_code folder if it doesn't exist
        if not source_code_folder:
            await self._studio_client.create_folder_async("source_code")
            yield FileOperationUpdate(
                file_path="source_code",
                status="created_folder",
                message="Created 'source_code' folder.",
            )
            source_code_files = {}

        # Get files to upload and process them
        files = files_to_include(
            settings,
            self.directory,
            self.include_uv_lock,
            directories_to_ignore=["evals"],
        )

        # Process all files and get updates (this includes HTTP calls for agent.json/entry-points.json)
        updates = await self._process_file_uploads(files, source_code_files, root_files)

        # Yield all updates
        for update in updates:
            yield update



"""Agent Models."""

from enum import Enum
from typing import Annotated, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Discriminator, Field, Tag, field_validator

from uipath.models import Connection
from uipath.models.guardrails import AgentEscalationRecipient, Guardrail


class AgentResourceType(str, Enum):
    """Enum for resource types."""

    TOOL = "tool"
    CONTEXT = "context"
    ESCALATION = "escalation"
    MCP = "mcp"


class BaseAgentResourceConfig(BaseModel):
    """Base resource model with common properties."""

    name: str
    description: str

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentUnknownResourceConfig(BaseAgentResourceConfig):
    """Fallback for unknown or future resource types."""

    resource_type: str = Field(alias="$resourceType")

    model_config = ConfigDict(extra="allow")


class BaseAgentToolResourceConfig(BaseAgentResourceConfig):
    """Tool resource with tool-specific properties."""

    resource_type: Literal[AgentResourceType.TOOL] = Field(alias="$resourceType")
    input_schema: Dict[str, Any] = Field(
        ..., alias="inputSchema", description="Input schema for the tool"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentToolType(str, Enum):
    """Agent tool type."""

    AGENT = "Agent"
    PROCESS = "Process"
    API = "Api"
    PROCESS_ORCHESTRATION = "ProcessOrchestration"
    INTEGRATION = "Integration"


class AgentToolSettings(BaseModel):
    """Settings for tool."""

    max_attempts: Optional[int] = Field(None, alias="maxAttempts")
    retry_delay: Optional[int] = Field(None, alias="retryDelay")
    timeout: Optional[int] = Field(None)

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class BaseResourceProperties(BaseModel):
    """Base resource properties."""

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentProcessToolProperties(BaseResourceProperties):
    """Properties specific to tool configuration."""

    folder_path: Optional[str] = Field(None, alias="folderPath")
    process_name: Optional[str] = Field(None, alias="processName")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentProcessToolResourceConfig(BaseAgentToolResourceConfig):
    """Tool resource with tool-specific properties."""

    type: Literal[
        AgentToolType.AGENT,
        AgentToolType.PROCESS,
        AgentToolType.API,
        AgentToolType.PROCESS_ORCHESTRATION,
    ]
    output_schema: Dict[str, Any] = Field(
        ..., alias="outputSchema", description="Output schema for the tool"
    )
    properties: AgentProcessToolProperties = Field(
        ..., description="Tool-specific properties"
    )
    settings: AgentToolSettings = Field(
        default_factory=AgentToolSettings, description="Tool settings"
    )
    arguments: Dict[str, Any] = Field(
        default_factory=dict, description="Tool arguments"
    )

    @field_validator("type", mode="before")
    @classmethod
    def normalize_type(cls, v: Any) -> str:
        """Normalize tool type from lowercase to properly cased enum values."""
        if isinstance(v, str):
            lowercase_mapping = {
                "agent": AgentToolType.AGENT,
                "process": AgentToolType.PROCESS,
                "api": AgentToolType.API,
                "processorchestration": AgentToolType.PROCESS_ORCHESTRATION,
            }
            return lowercase_mapping.get(v.lower(), v)
        return v

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentIntegrationToolParameter(BaseModel):
    """Agent integration tool parameter."""

    name: str = Field(..., alias="name")
    type: str = Field(..., alias="type")
    value: Optional[Any] = Field(None, alias="value")
    field_location: str = Field(..., alias="fieldLocation")

    # Useful Metadata
    display_name: Optional[str] = Field(None, alias="displayName")
    display_value: Optional[str] = Field(None, alias="displayValue")
    description: Optional[str] = Field(None, alias="description")
    position: Optional[str] = Field(None, alias="position")
    field_variant: Optional[str] = Field(None, alias="fieldVariant")
    dynamic: Optional[bool] = Field(None, alias="dynamic")
    is_cascading: Optional[bool] = Field(None, alias="isCascading")
    sort_order: Optional[int] = Field(..., alias="sortOrder")
    required: Optional[bool] = Field(None, alias="required")
    # enum_values, dynamic_behavior and reference not typed currently

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentIntegrationToolProperties(BaseResourceProperties):
    """Properties specific to tool."""

    tool_path: str = Field(..., alias="toolPath")
    object_name: str = Field(..., alias="objectName")
    tool_display_name: str = Field(..., alias="toolDisplayName")
    tool_description: str = Field(..., alias="toolDescription")
    method: str = Field(..., alias="method")
    connection: Connection = Field(..., alias="connection")
    body_structure: Optional[dict[str, Any]] = Field(None, alias="bodyStructure")
    parameters: List[AgentIntegrationToolParameter] = Field([], alias="parameters")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentIntegrationToolResourceConfig(BaseAgentToolResourceConfig):
    """Tool resource with tool-specific properties."""

    type: Literal[AgentToolType.INTEGRATION] = AgentToolType.INTEGRATION
    properties: AgentIntegrationToolProperties
    settings: Optional[AgentToolSettings] = Field(None, description="Tool settings")
    arguments: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Tool arguments"
    )
    is_enabled: Optional[bool] = Field(None, alias="isEnabled")

    @field_validator("type", mode="before")
    @classmethod
    def normalize_type(cls, v: Any) -> str:
        """Normalize tool type from lowercase to properly cased enum values."""
        if isinstance(v, str) and v.lower() == "integration":
            return AgentToolType.INTEGRATION
        return v

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentUnknownToolResourceConfig(BaseAgentResourceConfig):
    """Fallback for unknown or future tool types."""

    resource_type: Literal[AgentResourceType.TOOL] = AgentResourceType.TOOL
    type: str = Field(alias="$resourceType")
    arguments: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Tool arguments"
    )
    is_enabled: Optional[bool] = Field(None, alias="isEnabled")

    model_config = ConfigDict(extra="allow")


class AgentContextQuerySetting(BaseModel):
    """Query setting for context configuration."""

    description: Optional[str] = Field(None)
    variant: Optional[str] = Field(None)

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentContextValueSetting(BaseModel):
    """Generic value setting for context configuration."""

    value: Any = Field(...)

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentContextOutputColumn(BaseModel):
    """Output column configuration for Batch Transform."""

    name: str = Field(...)
    description: Optional[str] = Field(None)

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentContextSettings(BaseModel):
    """Settings for context."""

    result_count: int = Field(alias="resultCount")
    retrieval_mode: Literal["Semantic", "Structured", "DeepRAG", "BatchTransform"] = (
        Field(alias="retrievalMode")
    )
    threshold: float = Field(default=0)
    query: Optional[AgentContextQuerySetting] = Field(None)
    folder_path_prefix: Optional[Union[Dict[str, Any], AgentContextValueSetting]] = (
        Field(None, alias="folderPathPrefix")
    )
    file_extension: Optional[Union[Dict[str, Any], AgentContextValueSetting]] = Field(
        None, alias="fileExtension"
    )
    citation_mode: Optional[AgentContextValueSetting] = Field(
        None, alias="citationMode"
    )
    web_search_grounding: Optional[AgentContextValueSetting] = Field(
        None, alias="webSearchGrounding"
    )
    output_columns: Optional[List[AgentContextOutputColumn]] = Field(
        None, alias="outputColumns"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentContextResourceConfig(BaseAgentResourceConfig):
    """Context resource with context-specific properties."""

    resource_type: Literal[AgentResourceType.CONTEXT] = Field(alias="$resourceType")
    folder_path: str = Field(alias="folderPath")
    index_name: str = Field(alias="indexName")
    settings: AgentContextSettings = Field(..., description="Context settings")
    is_enabled: Optional[bool] = Field(None, alias="isEnabled")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentMcpTool(BaseModel):
    """MCP available tool."""

    name: str = Field(..., alias="name")
    description: str = Field(..., alias="description")
    input_schema: Dict[str, Any] = Field(..., alias="inputSchema")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentMcpResourceConfig(BaseAgentResourceConfig):
    """MCP resource configuration."""

    resource_type: Literal[AgentResourceType.MCP] = Field(alias="$resourceType")
    folder_path: str = Field(alias="folderPath")
    slug: str = Field(..., alias="slug")
    available_tools: List[AgentMcpTool] = Field(..., alias="availableTools")
    is_enabled: Optional[bool] = Field(None, alias="isEnabled")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentEscalationChannelProperties(BaseResourceProperties):
    """Agent escalation channel properties."""

    app_name: str = Field(..., alias="appName")
    app_version: int = Field(..., alias="appVersion")
    folder_name: Optional[str] = Field(..., alias="folderName")
    resource_key: str = Field(..., alias="resourceKey")
    is_actionable_message_enabled: Optional[bool] = Field(
        None, alias="isActionableMessageEnabled"
    )
    actionable_message_meta_data: Optional[Any] = Field(
        None, alias="actionableMessageMetaData"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentEscalationChannel(BaseModel):
    """Agent escalation channel."""

    id: Optional[str] = Field(None, alias="id")
    name: str = Field(..., alias="name")
    type: str = Field(alias="type")
    description: str = Field(..., alias="description")
    input_schema: Dict[str, Any] = Field(
        ..., alias="inputSchema", description="Input schema for the escalation channel"
    )
    output_schema: Dict[str, Any] = Field(
        ...,
        alias="outputSchema",
        description="Output schema for the escalation channel",
    )
    outcome_mapping: Optional[Dict[str, str]] = Field(None, alias="outcomeMapping")
    properties: AgentEscalationChannelProperties = Field(..., alias="properties")
    recipients: List[AgentEscalationRecipient] = Field(..., alias="recipients")
    task_title: Optional[str] = Field(default=None, alias="taskTitle")
    priority: Optional[str] = None
    labels: List[str] = Field(default_factory=list)

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentEscalationResourceConfig(BaseAgentResourceConfig):
    """Escalation resource with escalation-specific properties."""

    id: Optional[str] = Field(None, alias="id")
    resource_type: Literal[AgentResourceType.ESCALATION] = Field(alias="$resourceType")
    channels: List[AgentEscalationChannel] = Field(alias="channels")
    is_agent_memory_enabled: bool = Field(default=False, alias="isAgentMemoryEnabled")
    escalation_type: int = Field(default=0, alias="escalationType")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


def custom_discriminator(data: Any) -> str:
    """Discriminator for resource types. This is required due to multi-key discrimination requirements for resources."""
    if isinstance(data, dict):
        resource_type = data.get("$resourceType")
        if resource_type == AgentResourceType.CONTEXT:
            return "AgentContextResourceConfig"
        elif resource_type == AgentResourceType.ESCALATION:
            return "AgentEscalationResourceConfig"
        elif resource_type == AgentResourceType.MCP:
            return "AgentMcpResourceConfig"
        elif resource_type == AgentResourceType.TOOL:
            tool_type = data.get("type")
            if isinstance(tool_type, str):
                tool_type_lower = tool_type.lower()
                process_tool_types = {
                    AgentToolType.AGENT.value.lower(),
                    AgentToolType.PROCESS.value.lower(),
                    AgentToolType.API.value.lower(),
                    AgentToolType.PROCESS_ORCHESTRATION.value.lower(),
                }
                if tool_type_lower in process_tool_types:
                    return "AgentProcessToolResourceConfig"
                elif tool_type_lower == AgentToolType.INTEGRATION.value.lower():
                    return "AgentIntegrationToolResourceConfig"
                else:
                    return "AgentUnknownToolResourceConfig"
            else:
                return "AgentUnknownToolResourceConfig"
        else:
            return "AgentUnknownResourceConfig"
    raise ValueError("Invalid discriminator values")


AgentResourceConfig = Annotated[
    Union[
        Annotated[
            AgentProcessToolResourceConfig, Tag("AgentProcessToolResourceConfig")
        ],
        Annotated[
            AgentIntegrationToolResourceConfig,
            Tag("AgentIntegrationToolResourceConfig"),
        ],
        Annotated[
            AgentUnknownToolResourceConfig, Tag("AgentUnknownToolResourceConfig")
        ],
        Annotated[AgentContextResourceConfig, Tag("AgentContextResourceConfig")],
        Annotated[AgentEscalationResourceConfig, Tag("AgentEscalationResourceConfig")],
        Annotated[AgentMcpResourceConfig, Tag("AgentMcpResourceConfig")],
        Annotated[AgentUnknownResourceConfig, Tag("AgentUnknownResourceConfig")],
    ],
    Field(discriminator=Discriminator(custom_discriminator)),
]


class AgentMetadata(BaseModel):
    """Metadata for agent."""

    is_conversational: bool = Field(alias="isConversational")
    storage_version: str = Field(alias="storageVersion")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentMessageRole(str, Enum):
    """Enum for message roles."""

    SYSTEM = "system"
    USER = "user"


class AgentMessage(BaseModel):
    """Message model for agent definition."""

    role: Literal[AgentMessageRole.SYSTEM, AgentMessageRole.USER]
    content: str

    @field_validator("role", mode="before")
    @classmethod
    def normalize_role(cls, v: Any) -> str:
        """Normalize role to lowercase format."""
        if isinstance(v, str):
            return v.lower()
        return v

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentSettings(BaseModel):
    """Settings for agent."""

    engine: str = Field(..., description="Engine type, e.g., 'basic-v1'")
    model: str = Field(..., description="LLM model")
    max_tokens: int = Field(
        ..., alias="maxTokens", description="Maximum number of tokens per completion"
    )
    temperature: float = Field(..., description="Temperature for response generation")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class BaseAgentDefinition(BaseModel):
    """Agent definition model."""

    input_schema: Dict[str, Any] = Field(
        ..., alias="inputSchema", description="JSON schema for input arguments"
    )
    output_schema: Dict[str, Any] = Field(
        ..., alias="outputSchema", description="JSON schema for output arguments"
    )
    guardrails: Optional[List[Guardrail]] = Field(
        None, description="List of agent guardrails"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentType(str, Enum):
    """Agent type."""

    LOW_CODE = "lowCode"
    CODED = "coded"


class LowCodeAgentDefinition(BaseAgentDefinition):
    """Low code agent definition."""

    type: Literal[AgentType.LOW_CODE] = AgentType.LOW_CODE

    id: str = Field(..., description="Agent id or project name")
    name: str = Field(..., description="Agent name or project name")
    metadata: Optional[AgentMetadata] = Field(None, description="Agent metadata")
    messages: List[AgentMessage] = Field(
        ..., description="List of system and user messages"
    )

    version: str = Field("1.0.0", description="Agent version")
    resources: List[AgentResourceConfig] = Field(
        ..., description="List of tools, context, mcp and escalation resources"
    )
    features: List[Any] = Field(default_factory=list, description="Agent feature list")
    settings: AgentSettings = Field(..., description="Agent settings")

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class CodedAgentDefinition(BaseAgentDefinition):
    """Coded agent definition."""

    type: Literal[AgentType.CODED] = AgentType.CODED

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


KnownAgentDefinition = Annotated[
    Union[LowCodeAgentDefinition,],
    Field(discriminator="type"),
]


class UnknownAgentDefinition(BaseAgentDefinition):
    """Fallback for unknown agent definitions."""

    type: str

    model_config = ConfigDict(extra="allow")


AgentDefinition = Union[KnownAgentDefinition, UnknownAgentDefinition]



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, ConfigDict, Field


class FieldType(str, Enum):
    TEXT = "Text"
    NUMBER = "Number"
    DATE = "Date"
    NAME = "Name"
    ADDRESS = "Address"
    KEYWORD = "Keyword"
    SET = "Set"
    BOOLEAN = "Boolean"
    TABLE = "Table"
    INTERNAL = "Internal"


class ActionPriority(str, Enum):
    LOW = "Low"
    MEDIUM = "Medium"
    HIGH = "High"
    CRITICAL = "Critical"


class FieldValueProjection(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    id: str
    name: str
    value: str
    unformatted_value: str = Field(alias="unformattedValue")
    confidence: float
    ocr_confidence: float = Field(alias="ocrConfidence")
    type: FieldType


class FieldGroupValueProjection(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    field_group_name: str = Field(alias="fieldGroupName")
    field_values: List[FieldValueProjection] = Field(alias="fieldValues")


class ExtractionResult(BaseModel):
    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    document_id: str = Field(alias="DocumentId")
    results_version: int = Field(alias="ResultsVersion")
    results_document: dict = Field(alias="ResultsDocument")  # type: ignore
    extractor_payloads: Optional[List[dict]] = Field(  # type: ignore
        default=None, alias="ExtractorPayloads"
    )
    business_rules_results: Optional[List[dict]] = Field(  # type: ignore
        default=None, alias="BusinessRulesResults"
    )


class ExtractionResponse(BaseModel):
    """A model representing the response from a document extraction process.

    Attributes:
        extraction_result (ExtractionResult): The result of the extraction process.
        data_projection (List[FieldGroupValueProjection]): A simplified projection of the extracted data.
        project_id (str): The ID of the project associated with the extraction.
        tag (str): The tag associated with the published model version.
    """

    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    extraction_result: ExtractionResult = Field(alias="extractionResult")
    data_projection: List[FieldGroupValueProjection] = Field(alias="dataProjection")
    project_id: str = Field(alias="projectId")
    tag: str


class ValidationAction(BaseModel):
    """A model representing a validation action for a document.

    Attributes:
        action_data (dict): The data associated with the validation action.
        action_status (str): The status of the validation action. Possible values can be found in the [official documentation](https://docs.uipath.com/action-center/automation-cloud/latest/user-guide/about-actions#action-statuses).
        project_id (str): The ID of the project associated with the validation action.
        tag (str): The tag associated with the published model version.
        operation_id (str): The operation ID associated with the validation action.
    """

    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    action_data: dict = Field(alias="actionData")  # type: ignore
    action_status: str = Field(alias="actionStatus")
    project_id: str = Field(alias="projectId")
    tag: str
    operation_id: str = Field(alias="operationId")


class ValidatedResult(BaseModel):
    """A model representing the result of a validation action.

    Attributes:
        document_id (str): The ID of the validated document.
        results_document (dict): The validated results document.
    """

    model_config = ConfigDict(
        serialize_by_alias=True,
        validate_by_alias=True,
    )

    document_id: str = Field(alias="DocumentId")
    results_document: dict = Field(alias="ResultsDocument")  # type: ignore



import json
import os
import uuid
from typing import Any, Dict, Optional

from .._config import Config
from .._execution_context import ExecutionContext
from .._folder_context import FolderContext
from .._utils import Endpoint, RequestSpec, header_folder, infer_bindings
from .._utils.constants import ENV_JOB_KEY, HEADER_JOB_KEY
from ..models.job import Job
from ..tracing._traced import traced
from . import AttachmentsService
from ._base_service import BaseService


class ProcessesService(FolderContext, BaseService):
    """Service for managing and executing UiPath automation processes.

    Processes (also known as automations or workflows) are the core units of
    automation in UiPath, representing sequences of activities that perform
    specific business tasks.
    """

    def __init__(
        self,
        config: Config,
        execution_context: ExecutionContext,
        attachment_service: AttachmentsService,
    ) -> None:
        self._attachments_service = attachment_service
        super().__init__(config=config, execution_context=execution_context)

    @traced(name="processes_invoke", run_type="uipath")
    @infer_bindings(resource_type="process")
    def invoke(
        self,
        name: str,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Job:
        """Start execution of a process by its name.

        Related Activity: [Invoke Process](https://docs.uipath.com/activities/other/latest/workflow/invoke-process)

        Args:
            name (str): The name of the process to execute.
            input_arguments (Optional[Dict[str, Any]]): The input arguments to pass to the process.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.

        Returns:
            Job: The job execution details.

        Examples:
            ```python
            from uipath import UiPath

            client = UiPath()

            client.processes.invoke(name="MyProcess")
            ```

            ```python
            # if you want to execute the process in a specific folder
            # another one than the one set in the SDK config
            from uipath import UiPath

            client = UiPath()

            client.processes.invoke(name="MyProcess", folder_path="my-folder-key")
            ```
        """
        input_data = self._handle_input_arguments(
            input_arguments=input_arguments,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        spec = self._invoke_spec(
            name,
            input_data=input_data,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        response = self.request(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            content=spec.content,
            headers=spec.headers,
        )

        return Job.model_validate(response.json()["value"][0])

    @traced(name="processes_invoke", run_type="uipath")
    @infer_bindings(resource_type="process")
    async def invoke_async(
        self,
        name: str,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Job:
        """Asynchronously start execution of a process by its name.

        Related Activity: [Invoke Process](https://docs.uipath.com/activities/other/latest/workflow/invoke-process)

        Args:
            name (str): The name of the process to execute.
            input_arguments (Optional[Dict[str, Any]]): The input arguments to pass to the process.
            folder_key (Optional[str]): The key of the folder to execute the process in. Override the default one set in the SDK config.
            folder_path (Optional[str]): The path of the folder to execute the process in. Override the default one set in the SDK config.

        Returns:
            Job: The job execution details.

        Examples:
            ```python
            import asyncio

            from uipath import UiPath

            sdk = UiPath()

            async def main():
                job = await sdk.processes.invoke_async("testAppAction")
                print(job)

            asyncio.run(main())
            ```
        """
        input_data = await self._handle_input_arguments_async(
            input_arguments=input_arguments,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        spec = self._invoke_spec(
            name,
            input_data=input_data,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = await self.request_async(
            spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            content=spec.content,
            headers=spec.headers,
        )

        return Job.model_validate(response.json()["value"][0])

    @property
    def custom_headers(self) -> Dict[str, str]:
        return self.folder_headers

    def _handle_input_arguments(
        self,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Dict[str, str]:
        """Handle input arguments, storing as attachment if they exceed size limit.

        Args:
            input_arguments: The input arguments to process
            folder_key: The folder key for attachment storage
            folder_path: The folder path for attachment storage

        Returns:
            Dict containing either "InputArguments" or "InputFile" key
        """
        if not input_arguments:
            return {"InputArguments": json.dumps({})}

        # If payload exceeds limit, store as attachment
        payload_json = json.dumps(input_arguments)
        if len(payload_json) > 10000:  # 10k char limit
            attachment_id = self._attachments_service.upload(
                name=f"{uuid.uuid4()}.json",
                content=payload_json,
                folder_key=folder_key,
                folder_path=folder_path,
            )
            return {"InputFile": str(attachment_id)}
        else:
            return {"InputArguments": payload_json}

    async def _handle_input_arguments_async(
        self,
        input_arguments: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Dict[str, str]:
        """Handle input arguments, storing as attachment if they exceed size limit.

        Args:
            input_arguments: The input arguments to process
            folder_key: The folder key for attachment storage
            folder_path: The folder path for attachment storage

        Returns:
            Dict containing either "InputArguments" or "InputFile" key
        """
        if not input_arguments:
            return {"InputArguments": json.dumps({})}

        # If payload exceeds limit, store as attachment
        payload_json = json.dumps(input_arguments)
        if len(payload_json) > 10000:  # 10k char limit
            attachment_id = await self._attachments_service.upload_async(
                name=f"{uuid.uuid4()}.json",
                content=payload_json,
                folder_key=folder_key,
                folder_path=folder_path,
            )
            return {"InputFile": str(attachment_id)}
        else:
            return {"InputArguments": payload_json}

    def _invoke_spec(
        self,
        name: str,
        input_data: Optional[Dict[str, Any]] = None,
        *,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        request_spec = RequestSpec(
            method="POST",
            endpoint=Endpoint(
                "/orchestrator_/odata/Jobs/UiPath.Server.Configuration.OData.StartJobs"
            ),
            json={"startInfo": {"ReleaseName": name, **(input_data or {})}},
            headers={
                **header_folder(folder_key, folder_path),
            },
        )
        job_key = os.environ.get(ENV_JOB_KEY, None)
        if job_key:
            request_spec.headers[HEADER_JOB_KEY] = job_key

        return request_spec



import os
from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import uuid4

from rich.text import Text

from uipath.agent.conversation import UiPathConversationEvent, UiPathConversationMessage

from ...._runtime._contracts import UiPathErrorContract
from .._utils._chat import ConversationAggregator
from ._messages import LogMessage, TraceMessage


class ExecutionRun:
    """Represents a single execution run."""

    def __init__(
        self,
        entrypoint: str,
        input_data: Union[Dict[str, Any], UiPathConversationMessage],
        conversational: bool = False,
    ):
        self.id = str(uuid4())[:8]
        self.entrypoint = entrypoint
        self.input_data = input_data
        self.conversational = conversational
        self.resume_data: Optional[Any] = None
        self.output_data: Optional[Dict[str, Any]] = None
        self.start_time = datetime.now()
        self.end_time: Optional[datetime] = None
        self.status = "pending"  # pending, running, completed, failed, suspended
        self.traces: List[TraceMessage] = []
        self.logs: List[LogMessage] = []
        self.error: Optional[UiPathErrorContract] = None
        self.chat_aggregator = ConversationAggregator()

    @property
    def duration(self) -> str:
        if self.end_time:
            delta = self.end_time - self.start_time
            return f"{delta.total_seconds():.1f}s"
        elif self.start_time:
            delta = datetime.now() - self.start_time
            return f"{delta.total_seconds():.1f}s"
        return "0.0s"

    @property
    def display_name(self) -> Text:
        status_colors = {
            "pending": "grey50",
            "running": "yellow",
            "suspended": "cyan",
            "completed": "green",
            "failed": "red",
        }

        status_icon = {
            "pending": "●",
            "running": "▶",
            "suspended": "⏸",
            "completed": "✔",
            "failed": "✖",
        }.get(self.status, "?")

        script_name = (
            os.path.basename(self.entrypoint) if self.entrypoint else "untitled"
        )
        truncated_script = script_name[:8]
        time_str = self.start_time.strftime("%H:%M:%S")
        duration_str = self.duration[:6]

        text = Text()
        text.append(f"{status_icon:<2} ", style=status_colors.get(self.status, "white"))
        text.append(f"{truncated_script:<8} ")
        text.append(f"({time_str:<8}) ")
        text.append(f"[{duration_str:<6}]")

        return text

    @property
    def messages(self) -> List[UiPathConversationMessage]:
        return list(self.chat_aggregator.messages.values())

    def add_message(
        self, event: UiPathConversationEvent
    ) -> Optional[UiPathConversationMessage]:
        return self.chat_aggregator.add_event(event)



import logging
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from opentelemetry.sdk.trace import ReadableSpan
from pydantic import BaseModel, ConfigDict, Field, model_validator

from uipath._cli._evals._models._evaluation_set import EvaluationItem
from uipath.eval.models import EvalItemResult


class EvaluationEvents(str, Enum):
    CREATE_EVAL_SET_RUN = "create_eval_set_run"
    CREATE_EVAL_RUN = "create_eval_run"
    UPDATE_EVAL_SET_RUN = "update_eval_set_run"
    UPDATE_EVAL_RUN = "update_eval_run"


class EvalSetRunCreatedEvent(BaseModel):
    execution_id: str
    entrypoint: str
    eval_set_id: str
    no_of_evals: int
    evaluators: List[Any]


class EvalRunCreatedEvent(BaseModel):
    execution_id: str
    eval_item: EvaluationItem


class EvalItemExceptionDetails(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    runtime_exception: bool = False
    exception: Exception


class EvalRunUpdatedEvent(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    execution_id: str
    eval_item: EvaluationItem
    eval_results: List[EvalItemResult]
    success: bool
    agent_output: Any
    agent_execution_time: float
    spans: List[ReadableSpan]
    logs: List[logging.LogRecord]
    exception_details: Optional[EvalItemExceptionDetails] = None

    @model_validator(mode="after")
    def validate_exception_details(self):
        if not self.success and self.exception_details is None:
            raise ValueError("exception_details must be provided when success is False")
        return self


class EvalSetRunUpdatedEvent(BaseModel):
    execution_id: str
    evaluator_scores: dict[str, float]


ProgressEvent = Union[
    EvalSetRunCreatedEvent,
    EvalRunCreatedEvent,
    EvalRunUpdatedEvent,
    EvalSetRunUpdatedEvent,
]


class EventType(str, Enum):
    """Types of events that can be emitted during execution."""

    AGENT_MESSAGE = "agent_message"
    AGENT_STATE = "agent_state"
    ERROR = "error"


class UiPathRuntimeEvent(BaseModel):
    """Base class for all UiPath runtime events."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    event_type: EventType
    execution_id: Optional[str] = Field(
        default=None, description="The runtime execution id associated with the event"
    )
    metadata: Optional[Dict[str, Any]] = Field(
        default=None, description="Additional event context"
    )


class UiPathAgentMessageEvent(UiPathRuntimeEvent):
    """Event emitted when a message is created or streamed.

    Wraps framework-specific message objects (e.g., LangChain BaseMessage,
    CrewAI messages, AutoGen messages, etc.) without converting them.

    Attributes:
        payload: The framework-specific message object
        event_type: Automatically set to AGENT_MESSAGE
        metadata: Additional context

    Example:
        # LangChain
        event = UiPathAgentMessageEvent(
            payload=AIMessage(content="Hello"),
            metadata={"additional_prop": "123"}
        )

        # Access the message
        message = event.payload  # BaseMessage
        print(message.content)
    """

    payload: Any = Field(description="Framework-specific message object")
    event_type: EventType = Field(default=EventType.AGENT_MESSAGE, frozen=True)


class UiPathAgentStateEvent(UiPathRuntimeEvent):
    """Event emitted when agent state is updated.

    Wraps framework-specific state update objects, preserving the original
    structure and data from the framework.

    Attributes:
        payload: The framework-specific state update (e.g., LangGraph state dict)
        node_name: Name of the node/agent that produced this update (if available)
        event_type: Automatically set to AGENT_STATE
        metadata: Additional context

    Example:
        # LangGraph
        event = UiPathAgentStateEvent(
            payload={"messages": [...], "context": "..."},
            node_name="agent_node",
            metadata={"additional_prop": "123"}
        )

        # Access the state
        state = event.payload  # dict
        messages = state.get("messages", [])
    """

    payload: Dict[str, Any] = Field(description="Framework-specific state update")
    node_name: Optional[str] = Field(
        default=None, description="Name of the node/agent that caused this update"
    )
    event_type: EventType = Field(default=EventType.AGENT_STATE, frozen=True)



# type: ignore
"""CLI command for pulling remote project files from UiPath StudioWeb solution.

This module provides functionality to pull remote project files from a UiPath StudioWeb solution.
It handles:
- File downloads from source_code and evals folders
- Maintaining folder structure locally
- File comparison using hashes
- Interactive confirmation for overwriting files
"""

# type: ignore
import asyncio
import os
from pathlib import Path

import click

from ..telemetry import track
from ._utils._console import ConsoleLogger
from ._utils._constants import UIPATH_PROJECT_ID
from ._utils._project_files import ProjectPullError, pull_project

console = ConsoleLogger()


class InteractiveConflictHandler:
    """Handler that prompts user for each conflict."""

    def __init__(self, console: ConsoleLogger):
        self.console = console

    def should_overwrite(
        self, file_path: str, local_hash: str, remote_hash: str
    ) -> bool:
        self.console.warning(f" File {file_path} differs from remote version.")
        response = click.confirm("Do you want to overwrite it?", default=False)
        return response


@click.command()
@click.argument(
    "root",
    type=click.Path(exists=False, file_okay=False, dir_okay=True, path_type=Path),
    default=Path("."),
)
@track
def pull(root: Path) -> None:
    """Pull remote project files from Studio Web Project.

    This command pulls the remote project files from a UiPath Studio Web project.
    It downloads files from the source_code and evals folders, maintaining the
    folder structure locally. Files are compared using hashes before overwriting,
    and user confirmation is required for differing files.

    Args:
        root: The root directory to pull files into

    Environment Variables:
        UIPATH_PROJECT_ID: Required. The ID of the UiPath Studio Web project

    Example:
        $ uipath pull
        $ uipath pull /path/to/project
    """
    project_id = os.getenv(UIPATH_PROJECT_ID)
    if not project_id:
        console.error("UIPATH_PROJECT_ID environment variable not found.")

    default_download_configuration = {
        "source_code": root,
        "evals": root / "evals",
    }

    async def pull_with_updates():
        try:
            async for update in pull_project(
                project_id,
                default_download_configuration,
                InteractiveConflictHandler(console),
            ):
                console.info(update.message)
        except ProjectPullError as e:
            console.error(e.message, include_traceback=True)

    with console.spinner("Pulling UiPath project files..."):
        asyncio.run(pull_with_updates())



from enum import Enum
from typing import Annotated, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator


class FieldSource(str, Enum):
    """Field source enumeration."""

    INPUT = "input"
    OUTPUT = "output"


class ApplyTo(str, Enum):
    """Apply to enumeration."""

    INPUT = "input"
    INPUT_AND_OUTPUT = "inputAndOutput"
    OUTPUT = "output"


class FieldReference(BaseModel):
    """Field reference model."""

    path: str
    source: FieldSource

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class SelectorType(str, Enum):
    """Selector type enumeration."""

    ALL = "all"
    SPECIFIC = "specific"


class AllFieldsSelector(BaseModel):
    """All fields selector."""

    selector_type: Literal["all"] = Field(alias="$selectorType")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class SpecificFieldsSelector(BaseModel):
    """Specific fields selector."""

    selector_type: Literal["specific"] = Field(alias="$selectorType")
    fields: List[FieldReference]

    model_config = ConfigDict(populate_by_name=True, extra="allow")


FieldSelector = Annotated[
    Union[AllFieldsSelector, SpecificFieldsSelector],
    Field(discriminator="selector_type"),
]


class RuleType(str, Enum):
    """Rule type enumeration."""

    BOOLEAN = "boolean"
    NUMBER = "number"
    UNIVERSAL = "always"
    WORD = "word"


class WordOperator(str, Enum):
    """Word operator enumeration."""

    CONTAINS = "contains"
    DOES_NOT_CONTAIN = "doesNotContain"
    DOES_NOT_END_WITH = "doesNotEndWith"
    DOES_NOT_EQUAL = "doesNotEqual"
    DOES_NOT_START_WITH = "doesNotStartWith"
    ENDS_WITH = "endsWith"
    EQUALS = "equals"
    IS_EMPTY = "isEmpty"
    IS_NOT_EMPTY = "isNotEmpty"
    STARTS_WITH = "startsWith"


class WordRule(BaseModel):
    """Word rule model."""

    rule_type: Literal["word"] = Field(alias="$ruleType")
    field_selector: FieldSelector = Field(alias="fieldSelector")
    operator: WordOperator
    value: Optional[str] = None

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class UniversalRule(BaseModel):
    """Universal rule model."""

    rule_type: Literal["always"] = Field(alias="$ruleType")
    apply_to: ApplyTo = Field(alias="applyTo")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class NumberOperator(str, Enum):
    """Number operator enumeration."""

    DOES_NOT_EQUAL = "doesNotEqual"
    EQUALS = "equals"
    GREATER_THAN = "greaterThan"
    GREATER_THAN_OR_EQUAL = "greaterThanOrEqual"
    LESS_THAN = "lessThan"
    LESS_THAN_OR_EQUAL = "lessThanOrEqual"


class NumberRule(BaseModel):
    """Number rule model."""

    rule_type: Literal["number"] = Field(alias="$ruleType")
    field_selector: FieldSelector = Field(alias="fieldSelector")
    operator: NumberOperator
    value: float

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class BooleanOperator(str, Enum):
    """Boolean operator enumeration."""

    EQUALS = "equals"


class BooleanRule(BaseModel):
    """Boolean rule model."""

    rule_type: Literal["boolean"] = Field(alias="$ruleType")
    field_selector: FieldSelector = Field(alias="fieldSelector")
    operator: BooleanOperator
    value: bool

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class EnumListParameterValue(BaseModel):
    """Enum list parameter value."""

    parameter_type: Literal["enum-list"] = Field(alias="$parameterType")
    id: str
    value: List[str]

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class MapEnumParameterValue(BaseModel):
    """Map enum parameter value."""

    parameter_type: Literal["map-enum"] = Field(alias="$parameterType")
    id: str
    value: Dict[str, float]

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class NumberParameterValue(BaseModel):
    """Number parameter value."""

    parameter_type: Literal["number"] = Field(alias="$parameterType")
    id: str
    value: float

    model_config = ConfigDict(populate_by_name=True, extra="allow")


ValidatorParameter = Annotated[
    Union[EnumListParameterValue, MapEnumParameterValue, NumberParameterValue],
    Field(discriminator="parameter_type"),
]


Rule = Annotated[
    Union[WordRule, NumberRule, BooleanRule, UniversalRule],
    Field(discriminator="rule_type"),
]


class ActionType(str, Enum):
    """Action type enumeration."""

    BLOCK = "block"
    ESCALATE = "escalate"
    FILTER = "filter"
    LOG = "log"


class BlockAction(BaseModel):
    """Block action model."""

    action_type: Literal["block"] = Field(alias="$actionType")
    reason: str

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class FilterAction(BaseModel):
    """Filter action model."""

    action_type: Literal["filter"] = Field(alias="$actionType")
    fields: List[FieldReference]

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class SeverityLevel(str, Enum):
    """Severity level enumeration."""

    ERROR = "Error"
    INFO = "Info"
    WARNING = "Warning"


class LogAction(BaseModel):
    """Log action model."""

    action_type: Literal["log"] = Field(alias="$actionType")
    message: str = Field(..., alias="message")
    severity_level: SeverityLevel = Field(alias="severityLevel")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class EscalateActionApp(BaseModel):
    """Escalate action app model."""

    id: Optional[str] = None
    version: int
    name: str
    folder_id: Optional[str] = Field(None, alias="folderId")
    folder_name: str = Field(alias="folderName")
    app_process_key: Optional[str] = Field(None, alias="appProcessKey")
    runtime: Optional[str] = None

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentEscalationRecipientType(str, Enum):
    """Enum for escalation recipient types."""

    USER_ID = "UserId"
    GROUP_ID = "GroupId"
    USER_EMAIL = "UserEmail"


class AgentEscalationRecipient(BaseModel):
    """Recipient for escalation."""

    type: Union[AgentEscalationRecipientType, str] = Field(..., alias="type")
    value: str = Field(..., alias="value")
    display_name: Optional[str] = Field(default=None, alias="displayName")

    @field_validator("type", mode="before")
    @classmethod
    def normalize_type(cls, v: Any) -> str:
        """Normalize recipient type from int (1=UserId, 2=GroupId, 3=UserEmail) or string. Unknown integers are converted to string."""
        if isinstance(v, int):
            mapping = {
                1: AgentEscalationRecipientType.USER_ID,
                2: AgentEscalationRecipientType.GROUP_ID,
                3: AgentEscalationRecipientType.USER_EMAIL,
            }
            return mapping.get(v, str(v))
        return v

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class EscalateAction(BaseModel):
    """Escalate action model."""

    action_type: Literal["escalate"] = Field(alias="$actionType")
    app: EscalateActionApp
    recipient: AgentEscalationRecipient

    model_config = ConfigDict(populate_by_name=True, extra="allow")


GuardrailAction = Annotated[
    Union[BlockAction, FilterAction, LogAction, EscalateAction],
    Field(discriminator="action_type"),
]


class GuardrailScope(str, Enum):
    """Guardrail scope enumeration."""

    AGENT = "Agent"
    LLM = "Llm"
    TOOL = "Tool"


class GuardrailSelector(BaseModel):
    """Guardrail selector model."""

    scopes: List[GuardrailScope] = Field(default=[GuardrailScope.TOOL])
    match_names: Optional[List[str]] = Field(None, alias="matchNames")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class BaseGuardrail(BaseModel):
    """Base guardrail model."""

    id: str
    name: str
    description: Optional[str] = None
    action: GuardrailAction
    enabled_for_evals: bool = Field(True, alias="enabledForEvals")
    selector: GuardrailSelector

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class CustomGuardrail(BaseGuardrail):
    """Custom guardrail model."""

    guardrail_type: Literal["custom"] = Field(alias="$guardrailType")
    rules: List[Rule]

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class BuiltInValidatorGuardrail(BaseGuardrail):
    """Built-in validator guardrail model."""

    guardrail_type: Literal["builtInValidator"] = Field(alias="$guardrailType")
    validator_type: str = Field(alias="validatorType")
    validator_parameters: List[ValidatorParameter] = Field(
        default_factory=list, alias="validatorParameters"
    )

    model_config = ConfigDict(populate_by_name=True, extra="allow")


Guardrail = Annotated[
    Union[CustomGuardrail, BuiltInValidatorGuardrail],
    Field(discriminator="guardrail_type"),
]


class GuardrailType(str, Enum):
    """Guardrail type enumeration."""

    BUILT_IN_VALIDATOR = "builtInValidator"
    CUSTOM = "custom"


# Helper functions for type checking
def is_boolean_rule(rule: Rule) -> bool:
    """Check if rule is a BooleanRule."""
    return hasattr(rule, "rule_type") and rule.rule_type == RuleType.BOOLEAN


def is_number_rule(rule: Rule) -> bool:
    """Check if rule is a NumberRule."""
    return hasattr(rule, "rule_type") and rule.rule_type == RuleType.NUMBER


def is_universal_rule(rule: Rule) -> bool:
    """Check if rule is a UniversalRule."""
    return hasattr(rule, "rule_type") and rule.rule_type == RuleType.UNIVERSAL


def is_word_rule(rule: Rule) -> bool:
    """Check if rule is a WordRule."""
    return hasattr(rule, "rule_type") and rule.rule_type == RuleType.WORD


def is_custom_guardrail(guardrail: Guardrail) -> bool:
    """Check if guardrail is a CustomGuardrail."""
    return (
        hasattr(guardrail, "guardrail_type")
        and guardrail.guardrail_type == GuardrailType.CUSTOM
    )


def is_built_in_validator_guardrail(guardrail: Guardrail) -> bool:
    """Check if guardrail is a BuiltInValidatorGuardrail."""
    return (
        hasattr(guardrail, "guardrail_type")
        and guardrail.guardrail_type == GuardrailType.BUILT_IN_VALIDATOR
    )


def is_valid_action_type(value: Any) -> bool:
    """Check if value is a valid ActionType."""
    return isinstance(value, str) and value.lower() in [
        at.value.lower() for at in ActionType
    ]


def is_valid_severity_level(value: Any) -> bool:
    """Check if value is a valid SeverityLevel."""
    return isinstance(value, str) and value in [sl.value for sl in SeverityLevel]


# Guardrail Models
class AgentGuardrailRuleParameter(BaseModel):
    """Parameter for guardrail rules."""

    parameter_type: str = Field(..., alias="$parameterType")
    parameter_type_alt: Optional[str] = Field(None, alias="parameterType")
    value: Any = Field(..., description="Parameter value")
    id: str = Field(..., description="Parameter identifier")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailRule(BaseModel):
    """Guardrail validation rule."""

    rule_type: str = Field(..., alias="$ruleType")
    rule_type_alt: Optional[str] = Field(None, alias="ruleType")
    validator: str = Field(..., description="Validator type")
    parameters: List[AgentGuardrailRuleParameter] = Field(
        default_factory=list, description="Rule parameters"
    )

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailActionApp(BaseModel):
    """App configuration for guardrail actions."""

    name: str = Field(..., description="App name")
    version: str = Field(..., description="App version")
    folder_name: str = Field(..., alias="folderName", description="Folder name")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailActionRecipient(BaseModel):
    """Recipient for guardrail actions."""

    type: int = Field(..., description="Recipient type")
    value: str = Field(..., description="Recipient identifier")
    display_name: str = Field(..., alias="displayName", description="Display name")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailAction(BaseModel):
    """Action configuration for guardrails."""

    action_type: str = Field(..., alias="$actionType")
    action_type_alt: Optional[str] = Field(None, alias="actionType")
    app: AgentGuardrailActionApp = Field(..., description="App configuration")
    recipient: AgentGuardrailActionRecipient = Field(..., description="Recipient")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailSelector(BaseModel):
    """Selector for guardrail application scope."""

    scopes: List[str] = Field(..., description="Scopes where guardrail applies")
    match_names: List[str] = Field(
        ..., alias="matchNames", description="Names to match"
    )

    model_config = ConfigDict(populate_by_name=True, extra="allow")



import importlib.util
import inspect
import sys
from dataclasses import fields, is_dataclass
from enum import Enum
from types import ModuleType
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic import BaseModel

SchemaType = Literal["object", "integer", "number", "string", "boolean", "array"]

TYPE_MAP: Dict[str, SchemaType] = {
    "int": "integer",
    "float": "number",
    "str": "string",
    "bool": "boolean",
    "list": "array",
    "dict": "object",
    "List": "array",
    "Dict": "object",
}


def get_type_schema(type_hint: Any) -> Dict[str, Any]:
    """Convert a type hint to a JSON schema."""
    if type_hint is None or type_hint == inspect.Parameter.empty:
        return {"type": "object"}

    origin = get_origin(type_hint)
    args = get_args(type_hint)

    if origin is Union:
        if type(None) in args:
            real_type = next(arg for arg in args if arg is not type(None))
            return get_type_schema(real_type)
        return {"type": "object"}

    if origin in (list, List):
        item_type = args[0] if args else Any
        return {"type": "array", "items": get_type_schema(item_type)}

    if origin in (dict, Dict):
        return {"type": "object"}

    if inspect.isclass(type_hint):
        if issubclass(type_hint, Enum):
            enum_values = [member.value for member in type_hint]
            if not enum_values:
                return {"type": "string", "enum": []}

            first_value = enum_values[0]
            if isinstance(first_value, str):
                enum_type = "string"
            elif isinstance(first_value, int):
                enum_type = "integer"
            elif isinstance(first_value, float):
                enum_type = "number"
            elif isinstance(first_value, bool):
                enum_type = "boolean"
            else:
                enum_type = "string"

            return {"type": enum_type, "enum": enum_values}

        if issubclass(type_hint, BaseModel):
            properties = {}
            required = []

            # Get the model fields
            model_fields = type_hint.model_fields

            for field_name, field_info in model_fields.items():
                # Use alias if defined, otherwise use field name
                schema_field_name = field_info.alias if field_info.alias else field_name

                # Get the field type schema
                field_schema = get_type_schema(field_info.annotation)
                properties[schema_field_name] = field_schema

                # Check if field is required using Pydantic's built-in method
                if field_info.is_required():
                    required.append(schema_field_name)

            return {"type": "object", "properties": properties, "required": required}

        # Handle dataclasses
        elif is_dataclass(type_hint):
            properties = {}
            required = []

            for field in fields(type_hint):
                field_schema = get_type_schema(field.type)
                properties[field.name] = field_schema
                if field.default == field.default_factory:
                    required.append(field.name)

            return {"type": "object", "properties": properties, "required": required}

        # Handle regular classes with annotations
        elif hasattr(type_hint, "__annotations__"):
            properties = {}
            required = []

            for name, field_type in type_hint.__annotations__.items():
                field_schema = get_type_schema(field_type)
                properties[name] = field_schema
                # For regular classes, we'll consider all annotated fields as required
                # unless they have a default value in __init__
                if hasattr(type_hint, "__init__"):
                    sig = inspect.signature(type_hint.__init__)
                    if (
                        name in sig.parameters
                        and sig.parameters[name].default == inspect.Parameter.empty
                    ):
                        required.append(name)
                else:
                    required.append(name)

            return {"type": "object", "properties": properties, "required": required}

    type_name = type_hint.__name__ if hasattr(type_hint, "__name__") else str(type_hint)
    schema_type = TYPE_MAP.get(type_name, "object")

    return {"type": schema_type}


def load_module(file_path: str) -> ModuleType:
    """Load a Python module from file path."""
    spec = importlib.util.spec_from_file_location("dynamic_module", file_path)
    if not spec or not spec.loader:
        raise ImportError(f"Could not load spec for {file_path}")

    module = importlib.util.module_from_spec(spec)
    sys.modules["dynamic_module"] = module
    spec.loader.exec_module(module)
    return module


def generate_args(path: str) -> Dict[str, Dict[str, Any]]:
    """Generate input/output schema from main function type hints."""
    module = load_module(path)

    main_func = None
    for func_name in ["main", "run", "execute"]:
        if hasattr(module, func_name):
            main_func = getattr(module, func_name)
            break

    if not main_func:
        raise ValueError("No main function found in module")

    hints = get_type_hints(main_func)
    sig = inspect.signature(main_func)

    if not sig.parameters:
        return {"input": {}, "output": get_type_schema(hints.get("return", None))}

    input_param_name = next(iter(sig.parameters))
    input_schema = get_type_schema(hints.get(input_param_name))
    output_schema = get_type_schema(hints.get("return"))

    return {"input": input_schema, "output": output_schema}



import json
import os
from functools import wraps
from pathlib import PurePath
from typing import Any, Callable, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from uipath._utils.constants import HEADER_SW_LOCK_KEY
from uipath.models.exceptions import EnrichedException
from uipath.tracing import traced


class ProjectFile(BaseModel):
    """Model representing a file in a UiPath project.

    Attributes:
        id: The unique identifier of the file
        name: The name of the file
        is_main: Whether this is a main file
        file_type: The type of the file
        is_entry_point: Whether this is an entry point
        ignored_from_publish: Whether this file is ignored during publish
        app_form_id: The ID of the associated app form
        external_automation_id: The ID of the external automation
        test_case_id: The ID of the associated test case
    """

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )

    id: str = Field(alias="id")
    name: str = Field(alias="name")
    is_main: Optional[bool] = Field(default=None, alias="isMain")
    file_type: Optional[str] = Field(default=None, alias="fileType")
    is_entry_point: Optional[bool] = Field(default=None, alias="isEntryPoint")
    ignored_from_publish: Optional[bool] = Field(
        default=None, alias="ignoredFromPublish"
    )
    app_form_id: Optional[str] = Field(default=None, alias="appFormId")
    external_automation_id: Optional[str] = Field(
        default=None, alias="externalAutomationId"
    )
    test_case_id: Optional[str] = Field(default=None, alias="testCaseId")

    @field_validator("file_type", mode="before")
    @classmethod
    def convert_file_type(cls, v: Union[str, int, None]) -> Optional[str]:
        """Convert numeric file type to string.

        Args:
            v: The value to convert

        Returns:
            Optional[str]: The converted value or None
        """
        if isinstance(v, int):
            return str(v)
        return v


class ProjectFolder(BaseModel):
    """Model representing a folder in a UiPath project structure.

    Attributes:
        id: The unique identifier of the folder. Root folder id may be None.
        name: The name of the folder
        folders: List of subfolders
        files: List of files in the folder
        folder_type: The type of the folder
    """

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )

    id: Optional[str] = Field(default=None, alias="id")
    name: str = Field(alias="name")
    folders: List["ProjectFolder"] = Field(default_factory=list)
    files: List[ProjectFile] = Field(default_factory=list)
    folder_type: Optional[str] = Field(default=None, alias="folderType")

    @field_validator("folder_type", mode="before")
    @classmethod
    def convert_folder_type(cls, v: Union[str, int, None]) -> Optional[str]:
        """Convert numeric folder type to string.

        Args:
            v: The value to convert

        Returns:
            Optional[str]: The converted value or None
        """
        if isinstance(v, int):
            return str(v)
        return v


class ProjectStructure(ProjectFolder):
    """Model representing the complete file structure of a UiPath project.

    Attributes:
        id: The unique identifier of the root folder (optional)
        name: The name of the root folder (optional)
        folders: List of folders in the project
        files: List of files at the root level
        folder_type: The type of the root folder (optional)
    """

    pass


class LockInfo(BaseModel):
    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )
    project_lock_key: Optional[str] = Field(alias="projectLockKey")
    solution_lock_key: Optional[str] = Field(alias="solutionLockKey")


def get_folder_by_name(
    structure: ProjectStructure, folder_name: str
) -> Optional[ProjectFolder]:
    """Get a folder from the project structure by name.

    Args:
        structure: The project structure
        folder_name: Name of the folder to find

    Returns:
        Optional[ProjectFolder]: The found folder or None
    """
    for folder in structure.folders:
        if folder.name == folder_name:
            return folder
    return None


def resolve_path(
    folder: ProjectFolder,
    path: PurePath,
) -> ProjectFile | ProjectFolder:
    """Resolve a path relative to the folder.

    Args:
        folder: Project folder
        path: Path relative to the folder

    Returns: The resolved folder or file. If resolution fails, an assertion is raised.
    """
    root = path.parts
    while len(root) > 1:
        child = next(
            (folder for folder in folder.folders if folder.name == root[0]), None
        )
        assert child, "Path not found."
        folder = child
        root = root[1:]
    file = next((f for f in folder.files if f.name == root[0]), None)
    child = next((folder for folder in folder.folders if folder.name == root[0]), None)
    resolved = file or child
    assert resolved, "Path not found."
    return resolved


class AddedResource(BaseModel):
    """Represents a new file to be added during a structural migration."""

    content_file_path: Optional[str] = None
    parent_path: Optional[str] = None
    file_name: Optional[str] = None
    content_string: Optional[str] = None


class ModifiedResource(BaseModel):
    """Represents a file update during a structural migration."""

    id: str
    content_file_path: Optional[str] = None
    content_string: Optional[str] = None


class StructuralMigration(BaseModel):
    deleted_resources: List[str]
    added_resources: List[AddedResource]
    modified_resources: List[ModifiedResource]


def with_lock_retry(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        try:
            lock_info = await self._retrieve_lock()
            if not lock_info.project_lock_key:
                raise RuntimeError("Failed to retrieve project lock key.")

            headers = kwargs.get("headers", {}) or {}
            headers[HEADER_SW_LOCK_KEY] = lock_info.project_lock_key
            kwargs["headers"] = headers

            return await func(self, *args, **kwargs)
        except EnrichedException as e:
            if e.status_code == 423:
                from uipath._cli._utils._console import ConsoleLogger

                console = ConsoleLogger()
                console.error(
                    "The project is temporarily locked. This could be due to modifications or active processes. Please wait a moment and try again."
                )
            raise

    return wrapper


class StudioSolutionsClient:
    def __init__(self, solution_id: str):
        from uipath import UiPath

        self.uipath: UiPath = UiPath()
        self._solutions_base_url: str = f"/studio_/backend/api/Solution/{solution_id}"

    @traced(name="create_project", run_type="uipath")
    async def create_project_async(
        self,
        project_name: str,
        project_type: str = "Agent",
        trigger_type: str = "Manual",
    ):
        """Create a new project in the specified solution.

        Args:
            project_name: The name for the new project
            project_type: The type of project to create (default: "Agent")
            trigger_type: The trigger type for the project (default: "Manual")

        Returns:
            dict: The created project details including project ID
        """
        data = {
            "createDefaultProjectCommand[projectType]": project_type,
            "createDefaultProjectCommand[triggerType]": trigger_type,
            "createDefaultProjectCommand[name]": project_name,
        }

        response = await self.uipath.api_client.request_async(
            "POST",
            url=f"{self._solutions_base_url}/projects",
            data=data,
            scoped="org",
        )

        return response.json()


class StudioClient:
    def __init__(self, project_id: str):
        from uipath import UiPath

        self.uipath: UiPath = UiPath()
        self.file_operations_base_url: str = (
            f"/studio_/backend/api/Project/{project_id}/FileOperations"
        )
        self._lock_operations_base_url: str = (
            f"/studio_/backend/api/Project/{project_id}/Lock"
        )

    @traced(name="get_project_structure", run_type="uipath")
    async def get_project_structure_async(self) -> ProjectStructure:
        """Retrieve the project's file structure from UiPath Cloud.

        Makes an API call to fetch the complete file structure of a project,
        including all files and folders. The response is validated against
        the ProjectStructure model.

        Returns:
            ProjectStructure: The complete project structure
        """
        response = await self.uipath.api_client.request_async(
            "GET",
            url=f"{self.file_operations_base_url}/Structure",
            scoped="org",
        )

        return ProjectStructure.model_validate(response.json())

    @traced(name="create_folder", run_type="uipath")
    @with_lock_retry
    async def create_folder_async(
        self,
        folder_name: str,
        parent_id: Optional[str] = None,
        headers: Optional[dict[str, Any]] = None,
    ) -> str:
        """Create a folder in the project.

        Args:
            folder_name: Name of the folder to create
            parent_id: Optional parent folder ID
            headers: HTTP headers (automatically injected by decorator)

        Returns:
            str: The created folder ID
        """
        data = {"name": folder_name}
        if parent_id:
            data["parent_id"] = parent_id
        response = await self.uipath.api_client.request_async(
            "POST",
            url=f"{self.file_operations_base_url}/Folder",
            scoped="org",
            json=data,
            headers=headers or {},
        )
        return response.json()

    @traced(name="download_file", run_type="uipath")
    async def download_file_async(self, file_id: str) -> Any:
        response = await self.uipath.api_client.request_async(
            "GET",
            url=f"{self.file_operations_base_url}/File/{file_id}",
            scoped="org",
        )
        return response

    @traced(name="download_file", run_type="uipath")
    async def download_project_file_async(self, file: ProjectFile) -> Any:
        response = await self.uipath.api_client.request_async(
            "GET",
            url=f"{self.file_operations_base_url}/File/{file.id}",
            scoped="org",
        )
        return response

    @traced(name="upload_file", run_type="uipath")
    @with_lock_retry
    async def upload_file_async(
        self,
        *,
        local_file_path: Optional[str] = None,
        file_content: Optional[str] = None,
        file_name: str,
        folder: Optional[ProjectFolder] = None,
        remote_file: Optional[ProjectFile] = None,
        headers: Optional[dict[str, Any]] = None,
    ) -> tuple[str, str]:
        if local_file_path:
            with open(local_file_path, "rb") as f:
                file_content = f.read()  # type: ignore
        files_data = {"file": (file_name, file_content, "application/octet-stream")}

        if remote_file:
            # File exists in source_code folder, use PUT to update
            response = await self.uipath.api_client.request_async(
                "PUT",
                files=files_data,
                url=f"{self.file_operations_base_url}/File/{remote_file.id}",
                scoped="org",
                headers=headers or {},
            )
            action = "Updated"
        else:
            response = await self.uipath.api_client.request_async(
                "POST",
                url=f"{self.file_operations_base_url}/File",
                data={"parentId": folder.id} if folder else None,
                files=files_data,
                scoped="org",
                headers=headers or {},
            )
            action = "Uploaded"

        # response contains only the uploaded file identifier
        return response.json(), action

    @traced(name="delete_file", run_type="uipath")
    @with_lock_retry
    async def delete_item_async(
        self,
        item_id: str,
        headers: Optional[dict[str, Any]] = None,
    ) -> None:
        await self.uipath.api_client.request_async(
            "DELETE",
            url=f"{self.file_operations_base_url}/Delete/{item_id}",
            scoped="org",
            headers=headers or {},
        )

    def _resolve_content_and_filename(
        self,
        *,
        content_string: Optional[str],
        content_file_path: Optional[str],
        file_name: Optional[str] = None,
        modified: bool = False,
    ) -> tuple[bytes, Optional[str]]:
        """Resolve multipart content bytes and filename for a resource.

        Args:
            content_string: Inline content as a string.
            content_file_path: Path to a local file to read if inline content is not provided.
            file_name: Explicit filename to use when adding a new resource.

        Returns:
            A tuple of (content_bytes, filename).

        Raises:
            FileNotFoundError: If a provided file path does not exist.
            ValueError: If a filename cannot be determined.
        """
        content_bytes: bytes = b""
        resolved_name: Optional[str] = None
        if content_string is not None:
            content_bytes = content_string.encode("utf-8")
        elif content_file_path:
            if os.path.exists(content_file_path):
                with open(content_file_path, "rb") as f:
                    content_bytes = f.read()
            else:
                raise FileNotFoundError(f"File not found: {content_file_path}")

        if file_name:
            resolved_name = file_name
        elif content_file_path:
            resolved_name = os.path.basename(content_file_path)
        elif not modified:
            raise ValueError(
                "Unable to determine filename for multipart upload. "
                "When providing inline content (content_string), you must also provide file_name. "
                "Alternatively, set content_file_path so the filename can be inferred. "
                f"Received file_name={file_name!r}, content_file_path={content_file_path!r}."
            )

        return content_bytes, resolved_name

    @traced(name="synchronize_files", run_type="uipath")
    @with_lock_retry
    async def perform_structural_migration_async(
        self,
        structural_migration: StructuralMigration,
        headers: Optional[dict[str, Any]] = None,
    ) -> Any:
        """Perform structural migration of project files.

        Args:
            structural_migration: The structural migration data containing deleted and added resources
            headers: HTTP headers (automatically injected by decorator)

        Returns:
            Any: The API response
        """
        files: Any = []
        deleted_resources_json = json.dumps(structural_migration.deleted_resources)

        files.append(
            (
                "DeletedResources",
                (None, deleted_resources_json),
            )
        )
        for i, added_resource in enumerate(structural_migration.added_resources):
            content_bytes, filename = self._resolve_content_and_filename(
                content_string=added_resource.content_string,
                content_file_path=added_resource.content_file_path,
                file_name=added_resource.file_name,
            )

            files.append((f"AddedResources[{i}].Content", (filename, content_bytes)))

            if added_resource.parent_path:
                files.append(
                    (
                        f"AddedResources[{i}].ParentPath",
                        (None, added_resource.parent_path),
                    )
                )

        for i, modified_resource in enumerate(structural_migration.modified_resources):
            content_bytes, _ = self._resolve_content_and_filename(
                content_string=modified_resource.content_string,
                content_file_path=modified_resource.content_file_path,
                modified=True,
            )

            files.append((f"ModifiedResources[{i}].Content", content_bytes))
            files.append(
                (
                    f"ModifiedResources[{i}].Id",
                    (None, modified_resource.id),
                )
            )

        response = await self.uipath.api_client.request_async(
            "POST",
            url=f"{self.file_operations_base_url}/StructuralMigration",
            scoped="org",
            files=files,
            headers=headers or {},
        )

        return response

    async def _retrieve_lock(self) -> LockInfo:
        response = await self.uipath.api_client.request_async(
            "GET",
            url=f"{self._lock_operations_base_url}",
            scoped="org",
        )
        return LockInfo.model_validate(response.json())

    async def _put_lock(self):
        await self.uipath.api_client.request_async(
            "PUT",
            url=f"{self._lock_operations_base_url}/dummy-uuid-Shared?api-version=2",
            scoped="org",
        )



_CONNECTION_STRING = "InstrumentationKey=4128b5ed-f5ec-4612-ac32-b8cff0f2de93;IngestionEndpoint=https://northeurope-2.in.applicationinsights.azure.com/;LiveEndpoint=https://northeurope.livediagnostics.monitor.azure.com/;ApplicationId=f448014b-2abf-4681-8e48-583e33f7d173"

_APP_INSIGHTS_EVENT_MARKER_ATTRIBUTE = "APPLICATION_INSIGHTS_EVENT_MARKER_ATTRIBUTE"
_OTEL_RESOURCE_ATTRIBUTES = "OTEL_RESOURCE_ATTRIBUTES"
_SDK_VERSION = "SdkVersion"
_PROJECT_KEY = "ProjectKey"

_TELEMETRY_CONFIG_FILE = ".telemetry.json"

_CODE_FILEPATH = "code.file.path"
_CODE_FUNCTION = "code.function.name"
_CODE_LINENO = "code.line.number"

_CLOUD_ORG_ID = "CloudOrganizationId"
_CLOUD_TENANT_ID = "CloudTenantId"
_CLOUD_URL = "CloudUrl"
_APP_NAME = "ApplicationName"

_UNKNOWN = ""
